/* xlevel.c : Audio Level Meter for X11 (Linux and FreeBSD)
 * programmed by K. Takebe, 1997/11/02
 */
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <math.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Simple.h>

#include "xlevel.h"

/* geometry of level-meter */
#define left_x      (unit_margin)
#define left_y      8
#define right_x     (unit_margin)
#define right_y     24

/* global variables */
int debug   = 0;
int verbose = 0;
int nInd    = 0;
int nOver   = 0;
int nSpec   = 0;
int nDat    = 0;

/* only use in this source */
static char    dsp_device[80];
static char    mixer_device[80];
static char    mixing_devices[80]    = "all";
static char    available_devices[80];
static void    quitbykey();
XtCallbackProc redraw_window(Widget w, XEvent *event,
			     String *params, Cardinal *nparams);

int           unit_width;
int           unit_height;
int           unit_margin;
int           update;
int           peak;
int           over_level_ratio;
int           spectrum_mode;
int           usage_flag;
int           nolabel;
int           grid;

String        mixingDevices;
String        dspDevice;
String        mixerDevice;

unsigned int  color_under;
unsigned int  color_over;
unsigned int  color_none;
unsigned int  color_back;
static int    win_width, win_height;

/* for peak memory */
static int    old_left       = 0;
static int    old_right      = 0;

static int    *spec_l           = NULL, *spec_r           = NULL;
static int    *peak_array_l     = NULL, *peak_array_r     = NULL;
static int    *counter_array_l  = NULL, *counter_array_r  = NULL;
static int    *old_array_l      = NULL, *old_array_r      = NULL;
static int    *old_peak_array_l = NULL, *old_peak_array_r = NULL;

/* X valuable and resources */
XtAppContext         app_con;
XSetWindowAttributes atr;
XtTranslations	     newTranslations;

Widget        toplevel;
Widget        form;
Widget        frame;
Display      *display;
Window        window;
Cardinal      n;
Arg           args[10];

GC            gc0, gc1, gc2, gc3;

static XtResource	resources[] = {
  {"debugMode", "DebugMode", XtRBoolean, sizeof(Boolean),
     (Cardinal)&debug, XtRString, "False"},
  {"verboseMode", "VerboseMode", XtRBoolean, sizeof(Boolean),
     (Cardinal)&verbose, XtRString, "False"},
  {"usageFlag", "UsageFlag", XtRBoolean, sizeof(Boolean),
     (Cardinal)&usage_flag, XtRString, "False"},
  {"samplingRate", "SamplingRate", XtRInt, sizeof(int),
     (Cardinal)&sampling, XtRString, S_RATE},
  {"bufferSize", "BufferSize", XtRInt, sizeof(int),
     (Cardinal)&nDat, XtRString, N_DAT},
  {"dspDevice", "DspDevice", XtRString, sizeof(String),
     (Cardinal)&dspDevice, XtRString, DEFAULT_DSP_DEVICE},
  {"mixerDevice", "mixerDevice", XtRString, sizeof(String),
     (Cardinal)&mixerDevice, XtRString, DEFAULT_MIXER_DEVICE},
  {"indicatorNumber", "IndicatorNumber", XtRInt, sizeof(int),
     (Cardinal)&nInd, XtRString, N_IND},
  {"frequencyBands", "FrequencyBands", XtRInt, sizeof(int),
     (Cardinal)&nSpec, XtRString, N_SPEC},
  {"overRatio", "OverRatio", XtRInt, sizeof(int),
     (Cardinal)&over_level_ratio, XtRString, OVER_RATIO},
  {"update", "Update", XtRInt, sizeof(int),
     (Cardinal)&update, XtRString, UPDATE},
  {"devices", "Devices", XtRString, sizeof(String),
     (Cardinal)&mixingDevices, XtRString, "all"},
  {"bgColor", "bgColor", XtRPixel, sizeof(Pixel),
     (Cardinal)&color_back, XtRString, COLOR_BACK},
  {"indicatorNormal", "IndicatorNormal", XtRPixel, sizeof(Pixel),
     (Cardinal)&color_under, XtRString, COLOR_NORMAL},
  {"indicatorOver",   "IndicatorOver",   XtRPixel, sizeof(Pixel),
     (Cardinal)&color_over,  XtRString, COLOR_OVER},
  {"indicatorNone",   "IndicatorNone",   XtRPixel, sizeof(Pixel),
     (Cardinal)&color_none, XtRString,  COLOR_NONE},
  {"spectrumMode", "SpectrumMode", XtRBoolean, sizeof(Boolean),
     (Cardinal)&spectrum_mode, XtRString, "False"},
  {"indicatorWidth", "IndicatorWidth", XtRInt, sizeof(int),
     (Cardinal)&unit_width, XtRString,   UNIT_WIDTH},
  {"indicatorHeight", "IndicatorHeight", XtRInt, sizeof(int),
     (Cardinal)&unit_height, XtRString,  UNIT_HEIGHT},
  {"indicatorMargin", "IndicatorMargin", XtRInt, sizeof(int),
     (Cardinal)&unit_margin, XtRString,  UNIT_MARGIN},
  {"peak", "Peak", XtRInt, sizeof(int),
     (Cardinal)&peak, XtRString, N_PEAK},
  {"nolabel", "NoLabel", XtRBoolean, sizeof(Boolean),
     (Cardinal)&nolabel, XtRString, "False"},
  {"showGrid", "ShowGrid", XtRBoolean, sizeof(Boolean),
     (Cardinal)&grid, XtRString, "True"}
};

static XrmOptionDescRec options[] = {
  {"-debug",	   "*debugMode",       XrmoptionNoArg,  "true"},
  {"-v",	   "*verboseMode",     XrmoptionNoArg,  "true"},
  {"-verbose",	   "*verboseMode",     XrmoptionNoArg,  "true"},
  {"-help",	   "*usageFlag",       XrmoptionNoArg,  "true"},
  {"-dsp",	   "*dspDevice",       XrmoptionSepArg, NULL},
  {"-mixer",	   "*mixerDevice",     XrmoptionSepArg, NULL},
  {"-sf",	   "*samplingRate",    XrmoptionSepArg, NULL},
  {"-bufsize",     "*bufferSize",      XrmoptionSepArg, NULL},
  {"-dev",	   "*devices",         XrmoptionSepArg, NULL},
  {"-nodev",	   "*devices",         XrmoptionNoArg,  "none"},
  {"-update",	   "*update",          XrmoptionSepArg, NULL},
  {"-nind",        "*indicatorNumber", XrmoptionSepArg, NULL},
  {"-nband",       "*frequencyBands",  XrmoptionSepArg, NULL},
  {"-over",        "*overRatio",       XrmoptionSepArg, NULL},
  {"-backColor",   "*indicatorBack",   XrmoptionSepArg, NULL},
  {"-normalColor", "*indicatorNormal", XrmoptionSepArg, NULL},
  {"-overColor",   "*indicatorOver",   XrmoptionSepArg, NULL},
  {"-noneColor",   "*indicatorNone",   XrmoptionSepArg, NULL},
  {"-level",	   "*spectrumMode",    XrmoptionNoArg,  "false"},
  {"-spect",	   "*spectrumMode",    XrmoptionNoArg,  "true"},
  {"-grid",	   "*showGrid",        XrmoptionNoArg,  "true"},
  {"-nogrid",	   "*showGrid",        XrmoptionNoArg,  "false"},
  {"-peak",	   "*peak",	       XrmoptionSepArg, NULL},
  {"-nopeak",	   "*peak",	       XrmoptionNoArg,  "0"},
  {"-nolabel",	   "*nolabel",         XrmoptionNoArg,  "true"},
  {"-iw",	   "*indicatorWidth",  XrmoptionSepArg, NULL},
  {"-ih",	   "*indicatorHeight", XrmoptionSepArg, NULL},
  {"-im",	   "*indicatorMargin", XrmoptionSepArg, NULL}
};

static String fallback_resources[] = {
  "*level*font:    -*-helvetica-bold-r-normal--*-80-*-*-*-*-iso8859-1",
  "*spectrum*font: -*-helvetica-bold-o-normal--*-80-*-*-*-*-iso8859-1",
  "*foreground: White",
  "*background: Black",
  "*Form*defaultDistance: 0",
  "*Form*top: ChainTop",
  "*Form*bottom: ChainTop",
  "*Form*left: ChainLeft",
  "*Form*right: ChainLeft",
  "*level.label_L*label: L",
  "*level.label_R*label: R",
  "*level.label_R*fromVert: label_L",
  "*spectrum.label_L*label: Left",
  "*spectrum.label_R*label: Right",
  "*spectrum.label_R*fromHoriz: label_L",
  NULL
};

static char  *overrideTranslations = "\
	Ctrl<Key>Q:    quitbykey()\n\
	Ctrl<Key>R:     redraw_window()\n\
	<ResReq>:	redraw_window() \n\
	<Expose>:	redraw_window() \n";

static XtActionsRec actionTable[] = {
  {"redraw_window", (XtActionProc) redraw_window},
  {"quitbykey",     (XtActionProc) quitbykey},
  {NULL, (XtActionProc)NULL}
};


void print_usage(char *prog_name)
{
  printf("%s: Audio Level Meter for X11, ver.%s ", prog_name, XLEVEL_VERSION);
  printf("by take@bpel.tutics.tut.ac.jp\n");
  printf("available options\n");
  printf("  -help                : print this message\n");
  printf("  -v or -verbose       : verbose mode\n");
  printf("  -level               : normal level meter mode (default)\n");
  printf("  -spect               : spectrum-level meter mode\n");
  printf("  -sf          freq.   : sampling rate [Hz] (default: %s)\n",
	 S_RATE);
  printf("  -bufsize     size    : sampling buffer size (default: %s)\n",
	 N_DAT);
  printf("  -dsp         device  : dsp device (default: %s)\n",
	 DEFAULT_DSP_DEVICE);
  printf("  -mixer       device  : mixer device (default: %s)\n",
	 DEFAULT_MIXER_DEVICE);
  printf("  -update      miltime : update time (default: %s)\n",
	 UPDATE);
  printf("  -dev         devname : target devices (default: all; ");
  printf("ex. -dev Synth,Line)\n");
  check_available_devices(mixer_device, available_devices);
  printf("                         (avail. : %s)\n", available_devices);
  printf("  -nodev               : does not change target devices\n");
  printf("  -nind        number  : indicator number(default: %s)\n",
	 N_IND);
  printf("  -over  ratio         : max vs. over-level ratio(default: %s)\n",
	 OVER_RATIO);
  printf("  -peak num            : peak-level memory conter (default :%s)\n",
	 N_PEAK);
  printf("  -nopeak              : do not keep peak-level\n");
  printf("  -nolabel             : do not draw channel-labels\n");
  printf("  -backColor   cname   : background color(default: %s)\n",
	 COLOR_BACK);
  printf("  -normalColor cname   : normal indicator color(default: %s)\n",
	 COLOR_NORMAL);
  printf("  -overColor   cname   : over-level indicator color(default: %s)\n",
	 COLOR_OVER);
  printf("  -noneColor   cname   : color of inactive indicator(default: %s)\n",
	 COLOR_NONE);
  printf("  -iw width            : width of indicator (default: %s pixel)\n",
	 UNIT_WIDTH);
  printf("  -ih height           : height of indicator (default: %s pixel)\n",
	 UNIT_HEIGHT);
  printf("  -im margin           : margin of each indicator ");
  printf("(default: %s pixel)\n", UNIT_MARGIN);
  printf("\navailable SpectrumMode only:\n");
  printf("  -nband       number  : number of freq. band(default: %s)\n",
	 N_SPEC);
  printf("  -grid                : show grid at ");
  printf("500,1k,2k,5k,10k,20kHz(default)\n");
  printf("  -nogrid              : do not show grid\n");
  printf("\nother options are processed by X Window\n");
  printf("\n");
}

void
doquit()
{
  close_mixer_device();
  close_sound_device();
  exit(0);
}

void
destroy()
{
  doquit();
}

void quitbykey()
{
  doquit();
}


int init_spectrum_array()
{
  spec_l           = (int*)calloc(nSpec, sizeof(int));
  spec_r           = (int*)calloc(nSpec, sizeof(int));
  peak_array_l     = (int*)calloc(nSpec, sizeof(int));
  peak_array_r     = (int*)calloc(nSpec, sizeof(int));
  old_array_l      = (int*)calloc(nSpec, sizeof(int));
  old_array_r      = (int*)calloc(nSpec, sizeof(int));
  old_peak_array_l = (int*)calloc(nSpec, sizeof(int));
  old_peak_array_r = (int*)calloc(nSpec, sizeof(int));
  counter_array_l  = (int*)calloc(nSpec, sizeof(int));
  counter_array_r  = (int*)calloc(nSpec, sizeof(int));

  if (spec_l == NULL || peak_array_l == NULL ||
      old_array_l == NULL || counter_array_l == NULL ||
      spec_r == NULL || peak_array_r == NULL ||
      old_array_r == NULL || counter_array_r == NULL ) {
    perror("");
    doquit();
  }
  return 0;
}


int create_window()
{
  Widget label_L = NULL;

  if ( !spectrum_mode ) {
    form = XtCreateManagedWidget("level", formWidgetClass, toplevel,
				 NULL, ZERO);
    if ( !nolabel ) {
      label_L =
	XtCreateManagedWidget("label_L",labelWidgetClass,form,NULL,ZERO);
      XtCreateManagedWidget("label_R",labelWidgetClass,form,NULL,ZERO);
    }

    win_width  = (unit_width+unit_margin)*nInd+left_x+unit_height;
    win_height = right_y+unit_height+left_y;

    n = 0;
    XtSetArg(args[n], XtNwidth,  (Cardinal) win_width); n++;
    XtSetArg(args[n], XtNheight, (Cardinal) win_height); n++;
    if ( !nolabel )
      XtSetArg(args[n], XtNfromHoriz, label_L); n++;

    frame = XtCreateManagedWidget("frame", simpleWidgetClass, form,
				  args, n);
  } else {
    form = XtCreateManagedWidget("spectrum", formWidgetClass, toplevel,
				 NULL, ZERO);
    
    win_width  = unit_height*3+(unit_width+unit_margin)*nSpec*2-unit_margin*2;
    win_height = unit_height+(unit_height+unit_margin)*nInd;
    if (grid)
      win_height += unit_height+unit_margin;
    
    if ( !nolabel ) {
      n = 0;
      XtSetArg(args[n], XtNwidth,  (Cardinal) win_width/2); n++;
      label_L = 
	XtCreateManagedWidget("label_L", labelWidgetClass, form, args, n);
      XtCreateManagedWidget("label_R", labelWidgetClass, form, args, n);
    }

    n = 0;
    XtSetArg(args[n], XtNwidth,  (Cardinal) win_width); n++;
    XtSetArg(args[n], XtNheight, (Cardinal) win_height); n++;
    if ( !nolabel )
      XtSetArg(args[n], XtNfromVert, label_L); n++;
    frame = XtCreateManagedWidget("frame", simpleWidgetClass, form,
				  args, n);
  }
  XtRealizeWidget(toplevel);

  /* common settings */
  XtAppAddActions(app_con, actionTable, XtNumber(actionTable));
  newTranslations = XtParseTranslationTable(overrideTranslations);
  XtOverrideTranslations(form, newTranslations);

  window  = XtWindow(frame);

  gc0     = XCreateGC(display, window, 0, NULL);
  gc1     = XCreateGC(display, window, 0, NULL);
  gc2     = XCreateGC(display, window, 0, NULL);
  gc3     = XCreateGC(display, window, 0, NULL);
  XSetFillStyle(display, gc0, FillSolid);
  XSetFillStyle(display, gc1, FillSolid);
  XSetFillStyle(display, gc2, FillSolid);
  XSetFillStyle(display, gc3, FillSolid);
  XSetForeground(display, gc0, color_back);
  XSetForeground(display, gc1, color_under);
  XSetForeground(display, gc2, color_over);
  XSetForeground(display, gc3, color_none);

  atr.backing_store = Always;
  XChangeWindowAttributes(display, window, CWBackingStore, &atr);

  return 0;
}


int draw_horizontal(int level, int before, int pk, int pk_before, int x, int y)
{
  int          flag = 0;
  register int i;
  register int xpos; 
  int xstep = unit_width+unit_margin;

  if ( peak > 0 && level < pk_before )
    XFillRectangle(display, window, gc3,
		   x + pk_before*xstep, y, unit_width, unit_height);

  if ( before != level || before == 0 ) {
    if ( before < level ) {
      xpos = x+before*xstep; 
      for ( i = before; i <= level && i < nOver; i++, xpos += xstep )
        XFillRectangle(display, window, gc1, xpos, y, unit_width,unit_height);
      for ( ; i <= level && i < nInd; i++, xpos += xstep )
        XFillRectangle(display, window, gc2, xpos, y, unit_width,unit_height);
    } else {
      xpos = x+before*xstep;
      for ( i = before; i > level; i--, xpos -= xstep)
        XFillRectangle(display, window, gc3, xpos, y, unit_width,unit_height);
    }
    flag = 1;
  }
  if ( peak > 0 && level != pk && pk > 0 ) {
    XFillRectangle(display, window, (pk > nOver-1)? gc2 : gc1,
		   x + pk*xstep, y, unit_width, unit_height);
    flag = 1;
  }
  return flag;
}


int draw_vertical(int level, int before, int pk, int pk_before, int x, int y)
{
  int          flag = 0;
  register int i;
  register int ypos = 0; 
  int yorg  = y - unit_height;
  int ystep = -(unit_height+unit_margin);

  if (grid)
    yorg += ystep;

  if ( peak > 0 && level < pk_before )
    XFillRectangle(display, window, gc3,
		   x, yorg + pk_before*ystep, unit_width, unit_height);

  if ( before != level || before == 0 ) {
    if ( before < level ) {
      ypos = yorg+before*ystep; 
      for ( i = before; i <= level && i < nOver; i++, ypos += ystep )
        XFillRectangle(display, window, gc1, x, ypos, unit_width,unit_height);
      for ( ; i <= level && i < nInd; i++, ypos += ystep )
        XFillRectangle(display, window, gc2, x, ypos, unit_width,unit_height);
    } else {
      ypos = yorg+before*ystep;
      for ( i = before; i > level; i--, ypos -= ystep)
        XFillRectangle(display, window, gc3, x, ypos, unit_width,unit_height);
    }
    flag = 1;
  }
  if ( peak > 0 && level != pk && pk > 0 ) {
    XFillRectangle(display, window, (pk > nOver-1)? gc2 : gc1,
		   x, yorg + pk*ystep, unit_width, unit_height);
    flag = 1;
  }

  return flag;
}


int draw_level(int left, int right)
{
  static int    peak_left      = 0;
  static int    peak_right     = 0;
  static int    counter_left   = 0;
  static int    counter_right  = 0;
  static int    old_peak_left  = 0;
  static int    old_peak_right = 0;

  int flag;

  left  = (left  > nInd-1 ) ? nInd-1 :
    ((left  <= 0) ? ((old_left > 1)  ? old_left-1 : 0) : left);
  right = (right > nInd-1 ) ? nInd-1 :
    ((right <= 0) ? ((old_right > 1) ? old_right-1 : 0) : right);

  if ( peak > 0 ) {
    counter_left--;
    if ( peak_left < left || counter_left < 0 ) {
      old_peak_left  = peak_left;
      peak_left      = left;
      counter_left   = peak;
    }	  
    counter_right--;
    if ( peak_right < right || counter_right < 0 ) {
      old_peak_right = peak_right;
      peak_right     = right;
      counter_right  = peak;
    }	  
  }
  flag =  draw_horizontal(left, old_left,
			  peak_left, old_peak_left, left_x, left_y);
  flag += draw_horizontal(right,old_right,
			  peak_right,old_peak_right,right_x,right_y);
  old_left  = left;
  old_right = right;
  return flag;
}


int draw_spectrum(int left[], int right[])
{
  int flag;
  register int  i, a;
  int xpos, xstep, ypos;

  for ( i = 0; i < nSpec; i++ ) {
    a = left[i];   left[i] = (a > nInd-1 ) ? nInd-1 :
      ((a <= 0 ) ? ((old_array_l[i]>1) ? old_array_l[i]-1 : 0 ) : a );
    a = right[i]; right[i] = (a > nInd-1 ) ? nInd-1 :
      ((a <= 0 ) ? ((old_array_r[i]>1) ? old_array_r[i]-1 : 0 ) : a );
  }

  if ( peak > 0 ) {
    for ( i = 0; i < nSpec; i++ ) {
      counter_array_l[i]--;
      if ( peak_array_l[i] < left[i] || counter_array_l[i] < 0 ) {
	old_peak_array_l[i] = peak_array_l[i];
	peak_array_l[i] = left[i];
	counter_array_l[i] = peak;
      }	  
      counter_array_r[i]--;
      if ( peak_array_r[i] < right[i] || counter_array_r[i] < 0 ) {
	old_peak_array_r[i] = peak_array_r[i];
	peak_array_r[i] = right[i];
	counter_array_r[i] = peak;
      }	  
    }
  }

  flag = 0;
  ypos = win_height-unit_height;
  xstep = (unit_width+unit_margin);
  for ( xpos = unit_height, i = 0; i < nSpec; i++, xpos += xstep )
    flag += draw_vertical(left[i], old_array_l[i],
			  peak_array_l[i],old_peak_array_l[i], xpos, ypos);
  for ( xpos += unit_height, i = 0; i < nSpec; i++, xpos += xstep )
    flag += draw_vertical(right[i], old_array_r[i],
			  peak_array_r[i],old_peak_array_r[i], xpos, ypos);

  for ( i = 0; i < nSpec; i++ ) {
    old_array_l[i] = left[i];
    old_array_r[i] = right[i];
  }
  return flag;
}


inline void XLEVEL_EVENT_PROC() {
  XEvent event;
  while ( XtAppPending(app_con) != 0 ) {
    XtAppNextEvent(app_con, &event);
    XtDispatchEvent(&event);
  }
}

void redraw_level_window()
{
  int i, xpos;
  int xstep = unit_width+unit_margin;
  
  XFillRectangle(display, window, gc0, 0, 0, win_width, win_height);
  
  for ( xpos = left_x, i = 0; i < nInd; i++,  xpos += xstep )
    XFillRectangle(display,window,gc3,xpos,left_y,unit_width,unit_height);
  for ( xpos = right_x, i = 0; i < nInd; i++, xpos += xstep )
    XFillRectangle(display,window,gc3,xpos,right_y,unit_width,unit_height);

  XFillRectangle(display,window,gc1,left_x, left_y,unit_width,unit_height);
  XFillRectangle(display,window,gc1,right_x,right_y,unit_width,unit_height);
  
  old_left  = 0;
  old_right = 0;
}


void redraw_spectrum_window()
{
  int i, j, xpos, ypos = 0;
  int xstep = unit_width +unit_margin;
  int ystep = unit_height+unit_margin;
  
  XFillRectangle(display, window, gc0, 0, 0, win_width, win_height);
  
  for ( xpos = unit_height, i = 0; i < nSpec; i++,  xpos += xstep )
    for ( ypos = unit_margin, j = 0; j < nInd-1; j++, ypos += ystep )
      XFillRectangle(display,window,gc3,xpos,ypos,unit_width,unit_height);

  XLEVEL_EVENT_PROC();
  XFlush(display);

  for ( xpos += unit_height, i = 0; i < nSpec; i++,  xpos += xstep )
    for ( ypos = unit_margin, j = 0; j < nInd-1; j++, ypos += ystep )
      XFillRectangle(display,window,gc3,xpos,ypos,unit_width,unit_height);

  XLEVEL_EVENT_PROC();
  XFlush(display);

  for ( xpos = unit_height, i = 0; i < nSpec; i++,  xpos += xstep )
      XFillRectangle(display,window,gc1,xpos,ypos,unit_width,unit_height);
  for ( xpos += unit_height, i = 0; i < nSpec; i++,  xpos += xstep )
      XFillRectangle(display,window,gc1,xpos,ypos,unit_width,unit_height);

  if (grid) {
    int pos;
    int table1[2] = { 1000, 10000};
    int table2[4] = { 500, 2000, 5000, 20000};
#define SPEC_POS(x) ((double)nSpec*(1.0+log10(2.0*(double)(x)/(double)sampling)/2.0))

    ypos += ystep;

    for ( i = 0; i < 2; i++ ) {
      if ( table1[i] <= sampling/2 ) {
	if (( pos = xstep*SPEC_POS(table1[i]) ) >= 0 ) {
	  xpos  = unit_height + pos;
	  XFillRectangle(display,window,gc2,xpos,ypos,unit_margin,unit_height);
	  
	  xpos  = unit_height*2 + xstep*nSpec + pos;
	  XFillRectangle(display,window,gc2,xpos,ypos,unit_margin,unit_height);
	}
      }
    }
    for ( i = 0; i < 4; i++ ) {
      if ( table2[i] <= sampling/2 ) {
	if (( pos = xstep*SPEC_POS(table2[i]) ) >= 0 ) {
	  xpos  = unit_height + pos;
	  XFillRectangle(display,window,gc2,xpos,ypos,unit_margin,unit_margin);

	  xpos  = unit_height*2 + xstep*nSpec + pos;
	  XFillRectangle(display,window,gc2,xpos,ypos,unit_margin,unit_margin);
	}
      }
    }
  }

  XLEVEL_EVENT_PROC();
  XFlush(display);

  for ( i = 0; i < nSpec; i++ ) {
    old_array_l[i] = 0;
    old_array_r[i] = 0;
  }
}


XtCallbackProc redraw_window(w, event, params, nparams)
     Widget	 w;
     XEvent	*event;
     String     *params;
     Cardinal   *nparams;
{
  if (debug)
    printf("redraw_window(): redraw event\n");

  if ( !spectrum_mode )
    redraw_level_window();
  else
    redraw_spectrum_window();

  XFlush(display);
  return NULL;
}


void draw_level_loop()
{
  int i, left = 0, right = 0;
  int spend_time, increasing;

  if (verbose||debug)
    printf("\n");

  while(1) {
    store_start_time();
    
    get_level(&left, &right);
    if ( draw_level(left, right) > 0 )
      XFlush(display);
    
    XLEVEL_EVENT_PROC();
    
    increasing = 0;
    spend_time = get_spend_time();
    for ( i = 0; i < 1000 && update > spend_time+increasing; i++ ) {
      store_level_data();
      increasing  = get_spend_time() - spend_time;
      spend_time += increasing;
    }
    if (verbose||debug) {
      printf("\r\tactual update time = %5d msec", spend_time);
      fflush(stdout);
    }
  }
}

void draw_spectrum_loop()
{
  int i;
  int spend_time, increasing;

  if (verbose||debug)
    printf("\n");

  while(1) {
    store_start_time();
    
    get_spectrum(spec_l, spec_r);
    if ( draw_spectrum(spec_l, spec_r) > 0 )
      XFlush(display);
    
    XLEVEL_EVENT_PROC();
    
    increasing = 0;
    spend_time = get_spend_time();
    for ( i = 0; i < 1000 && update > spend_time+increasing; i++ ) {
      store_spectrum_data();
      increasing  = get_spend_time() - spend_time;
      spend_time += increasing;
    }
    if (verbose||debug) {
      printf("\r\tactual update time = %5d msec", spend_time);
      fflush(stdout);
    }
  }
}


int set_and_check_parameters() {
  strcpy(mixing_devices, mixingDevices);
  strcpy(dsp_device,     dspDevice);
  strcpy(mixer_device,   mixerDevice);

  if ( nDat < NDAT_MIN ) nDat = NDAT_MIN;
  if ( nDat > NDAT_MAX ) nDat = NDAT_MAX;

  if ( nInd  < MIN_INDICATORS ) nInd  = MIN_INDICATORS;
  if ( nInd  > MAX_INDICATORS ) nInd  = MAX_INDICATORS;
  if ( nSpec < MIN_FREQBANDS  ) nSpec = MIN_FREQBANDS;
  if ( nSpec > MAX_FREQBANDS  ) nSpec = MAX_FREQBANDS;
  if ( over_level_ratio > 0 && over_level_ratio < 100 )
    nOver = nInd * over_level_ratio / 100;
  if ( over_level_ratio <= 0  ) nOver = 0;
  if ( over_level_ratio > 100 ) nOver = nInd;

  if ( update <= 0 )      update       = 1;
  if ( unit_width  <= 0  ) unit_width  = 1;
  if ( unit_width  >  50 ) unit_width  = 50;
  if ( unit_height <= 0  ) unit_height = 1;
  if ( unit_height >  50 ) unit_height = 50;
  if ( unit_margin <  0  ) unit_margin = 0;
  if ( unit_margin >  25 ) unit_margin = 25;

  if (debug||verbose) {
    printf("\trunning %s mode\n",
	   (spectrum_mode)? "Spectrum" : "Normal");
    printf("\tupdate interval    = %d msec.\n", update);
    if ( peak > 0 )
      printf("\tpeak-level         = kept while %d updates\n", peak);
    printf("\tindicators         = %d\n", nInd);
    printf("\tover-level start   = %d\n", nOver);
    printf("\tfreq.-bands        = %d\n", nSpec);
    printf("\tbuffer size        = %d\n", nDat);
  }
  return 0;
}

int main(int argc, char *argv[])
{
  signal(SIGHUP,  (void (*)()) destroy);
  signal(SIGBUS,  (void (*)()) destroy);
  signal(SIGSEGV, (void (*)()) destroy);
  signal(SIGFPE,  (void (*)()) destroy);

  display  = XOpenDisplay(NULL);
  toplevel = XtAppInitialize(&app_con, "XLevel",
			     options, XtNumber(options), &argc, argv,
			     fallback_resources, NULL, ZERO);
    
  XtGetApplicationResources(toplevel, (caddr_t) NULL,
			    resources, XtNumber(resources), NULL, 0);

  if (verbose)
    printf("xlevel : Audio Level Meter for X11 on Linux/FreeBSD ver.%s\n",
	   XLEVEL_VERSION);
  if (debug)
    printf("xlevel : debug Mode\n");

  set_and_check_parameters();

  if ( usage_flag ) {
    print_usage(argv[0]);
    doquit();
  }

  init_sampling_buffer();  
  init_sound_card(dsp_device);
  mixer_setting(mixer_device, mixing_devices);

  init_level_trans_table();

  create_window();

  if ( !spectrum_mode ) {
    redraw_level_window();
    XLEVEL_EVENT_PROC();
    XFlush(display);
    draw_level_loop();
  } else {
    init_spectrum_array();
    init_spectrum_buffer();
    XLEVEL_EVENT_PROC();
    XFlush(display);
    redraw_spectrum_window();
    draw_spectrum_loop();
  }

  return 0;
}
